#define _POSIX_C_SOURCE 200112L /* needed for struct addrinfo */

#include "db.h"
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>

typedef struct {
    char *data_dir;
    char *log_dir;
    char *tmp_dir;
    char *gallery_dl_conf;
    char *yt_dlp_conf;
} Config;

int h_config_parser(Config *config) {
    char *config_location = malloc(10000);
    char *user_name = getenv("USER");
    if (user_name == NULL) {
        fprintf(stderr, "Couldn't read env variable USER\n");
        return 1;
    }
    sprintf(config_location, "/home/%s/.config/Hentai.conf", user_name);
    
    FILE *fptr = fopen(config_location, "r");
    if (fptr == NULL) {
        fprintf(stderr, "Failed to open the config %s\n", config_location);
        return 1;
    }

    char *line = malloc(10000);
    char *left = malloc(10000);
    char *right = malloc(10000);

    while(fgets(line, 10000, fptr) != NULL) {
        line[strcspn(line, "\n#")] = '\0';
        if (strcmp(line, "") == 0) {
            continue;
        }
        if (
            strchr(line, '=') == NULL
            || line[strcspn(line, "=") + 1] == ' '
            || line[strcspn(line, "=") - 1] == ' '
        ) {
            fprintf(stderr, "Config line doesn't contain a valid assignment. line = '%s'\n", line);
            continue;
        }
        /* TODO: support env variables */
        if (strchr(line, '$') != NULL) {
            fprintf(stderr, "Config doesn't support env variables. line = '%s'\n", line);
            continue;
        }
        
        left = strtok(line, "=");
        right = strtok(NULL, "\0");
        if (strcmp(left, "data_dir") == 0) {
            config->data_dir = malloc(10000);
            memcpy(config->data_dir, right, strlen(right));
        } else if (strcmp(left, "log_dir") == 0) {
            config->log_dir = malloc(10000);
            memcpy(config->log_dir, right, strlen(right));
        } else if (strcmp(left, "tmp_dir") == 0) {
            config->tmp_dir = malloc(10000);
            memcpy(config->tmp_dir, right, strlen(right));
        } else if (strcmp(left, "gallery-dl_conf") == 0) {
            config->gallery_dl_conf = malloc(10000);
            memcpy(config->gallery_dl_conf, right, strlen(right));
        } else if (strcmp(left, "yt-dlp_conf") == 0) {
            config->yt_dlp_conf = malloc(10000);
            memcpy(config->yt_dlp_conf, right, strlen(right));
        } else {
            fprintf(stderr, "Unknown config option. left = '%s'\n", left);
        }
    }
    fclose(fptr);
    free(line);
    free(config_location);
    return 0;
}

int h_command_builder(char *command, char **args) {
    char *tmp_command = malloc(strlen(command));
    memcpy(tmp_command, command, strlen(command));
    char *tmp_token;
    size_t index = 0;

    if (tmp_command == NULL) { return 1; }

    tmp_token = strtok(tmp_command, " ");
    while(tmp_token != NULL) {
        args[index] = malloc(strlen(tmp_token));
        memcpy(args[index], tmp_token, strlen(tmp_token));
        tmp_token = strtok(NULL, " ");
        index++;
    }
    free(tmp_command);
    args[index++] = NULL;
    return 0;
}

int h_run_cmd_and_wait(char *command) {
    char **args = malloc(strlen(command));
    int ret = h_command_builder(command, args);
    if (ret != 0) {
        return 1;
    }

    pid_t pid = fork();
    int status;

    if (pid == -1) {
        printf("Couldn't fork'\n");
        return 1;
    } else if (pid > 0) {
        waitpid(pid, &status, 0);
    } else {
        execvp(args[0], args);
    }
    free(args);

    return status;
}

int h_has_internet() {
    int ret;
    
    /* https://www.man7.org/linux/man-pages/man3/getaddrinfo.3.html */
    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    struct addrinfo *res, *rp;
    ret = getaddrinfo("hopeless-cloud.xyz", "80", 0, &res);
    if (ret != 0) {
        return 1;
    }
    int sockfd;
    for (rp = res; rp != NULL; rp = rp->ai_next) {
        /* https://www.man7.org/linux/man-pages/man2/socket.2.html */
        sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
        if (sockfd == -1) { continue; }
        
        /* https://www.man7.org/linux/man-pages/man2/connect.2.html */
        if (connect(sockfd, rp->ai_addr, rp->ai_addrlen) == 0) { break; }
        
        close(sockfd);
    }
    freeaddrinfo(res);
    if (rp == NULL) {
        return 1;
    }
    
    /* https://www.man7.org/linux/man-pages/man2/write.2.html */
    char *req_msg = "GET / HTTP/1.1\r\n\r\n";
    if (write(sockfd, req_msg, strlen(req_msg)) != strlen(req_msg)) {
        return 1;
    }
    
    /* https://www.man7.org/linux/man-pages/man2/recv.2.html */
    size_t count = 1000000;
    char buf[count];
    if (recv(sockfd, buf, count-1, MSG_WAITALL) == -1 ) {
        return 1;
    }
    close(sockfd);
    return 0;
}

int h_download(char *db_row, Config config) {
    int ret;
    char *url = malloc(strlen(db_row));
    char *artist_name = malloc(strlen(db_row));
    char *website_name = malloc(strlen(db_row));
    sscanf(db_row, "%s %s %s", url, artist_name, website_name);
    char *path = malloc(strlen(artist_name) + strlen(website_name) + 2);
    sprintf(path, "%s/%s", artist_name, website_name);
    char *command = malloc(10000);
    if (strstr(website_name, "Iwara") == 0) {
        sprintf(command, "yt-dlp_linux --config-locations %s -o %s/%(id)s.%(ext)s %s", config.yt_dlp_conf, path, url);
    } else {
        sprintf(command, "gallery-dl --config %s --directory %s/%s %s", config.gallery_dl_conf, config.data_dir, path, url);
    }
    free(url);
    free(artist_name);
    free(website_name);

    size_t i = 5;
    for (; i > 0; i--) {
        printf("%s\n", command);
        /* h_run_cmd_and_wait(command) */
        ret = h_has_internet();
        if (ret != 0) {
            fprintf(stderr, "No internet\n");
            return 1;
        }
    }
    free(command);
    return 0;
}

int main(int argc, char **argv) {
    int ret;
    Config config = { 0 };
    
    ret = h_config_parser(&config);
    if (ret != 0) {
        fprintf(stderr, "Failed to parse config\n");
        return 1;
    }

    if (argc < 2) {
        fprintf(stderr, "Incorrect amount of arguments\n");
        return 1;
    }

    ret = h_db_open(":memory:");
    if (ret != 0) {
        return 1;
    }

    if (strcmp(argv[1], "add") == 0) {
        char *db_row = malloc(strlen(argv[2]) * 2);
        memset(db_row, 0, strlen(argv[2]));

        if (argc == 3) {
            ret = h_db_add(db_row, argv[2]);
            if (ret != 0) {
                h_db_close();
                return 1;
            } 
        } else if (argc == 4) {
            ret = h_db_add_custom_name(db_row, argv[2], argv[3]);
            if (ret != 0) {
                h_db_close();
                return 1;
            }
        } else {
            h_db_close();
            fprintf(stderr, "Incorrect amount of arguments\n");
            return 1;
        }
        ret = h_download(db_row, config);
        free(db_row);
        if (ret != 0) {
            h_db_close();
            return 1;
        }
    } else if (strcmp(argv[1], "download_website") == 0) {
        char *website_name = argv[2];
        char *file_name = malloc(10000);
        if (argc == 3) {
            sprintf(file_name, "%s_generated.txt", website_name);
            FILE *fptr = fopen(file_name, "w+");
            free(file_name);
            if (fptr == NULL) {
                fprintf(stderr, "Couldn't open file\n");
                h_db_close();
                return 1;
            }

            ret = h_db_generate_file(fptr, argv[2]);
            fclose(fptr);
            if (ret != 0) {
                h_db_close();
                return 1;
            }
        } else if (argc == 4) {
            sprintf(file_name, "%s/%s_generated.txt", argv[3], website_name);
            FILE *fptr = fopen(file_name, "w+");
            free(file_name);
            if (fptr == NULL) {
                fprintf(stderr, "Couldn't open file\n");
                h_db_close();
                return 1;
            }

            ret = h_db_generate_file(fptr, argv[2]);
            fclose(fptr);
            if (ret != 0) {
                h_db_close();
                return 1;
            }
        } else {
            fprintf(stderr, "Incorrect amount of arguments\n");
            h_db_close();
            return 1;
        }
    } else {
        fprintf(stderr, "Invalid argument\n");
        h_db_close();
        return 1;
    }

    h_db_close();
    return 0;
}
